"""
Likely the format going forward, but kept separate for compatibility.
Still need to adapt eloqua_utils to use a similar output format as salesforce_utils.
Still need to add/finalize external validation toolset.
"""

import logging
from nubium_utils.dict_manipulator import DictManipulator
import json
import dictdiffer
import pytest

LOGGER = logging.getLogger(__name__)


def remove_nu_null(flat_dict):
    return {k: v for k, v in flat_dict.items() if '--NU--NULL' not in v}


def unflatten(flat_dict):
    return DictManipulator(flat_dict).to_dict()


def fill_schema_dict(flat_dict, empty_schema_dict):
    filled_dict = DictManipulator(empty_schema_dict)
    for k, v in flat_dict.items():
        filled_dict[k] = v
    return filled_dict.to_dict()


def nest_sort(d):
    if isinstance(d, dict):
        keys = sorted(d.keys())
        return {k: nest_sort(d[k]) for k in keys}
    elif isinstance(d, list):
        return [nest_sort(i) for i in d]
    else:
        return d


@pytest.fixture()
def preprocess_external_create_data(external_create_data_csv):
    return [remove_nu_null(rec) for rec in external_create_data_csv]


@pytest.fixture()
def preprocess_external_update_data(external_update_data_csv):
    return [remove_nu_null(rec) for rec in external_update_data_csv]


@pytest.fixture()
def preprocess_external_delete_data(external_delete_data_csv):
    return [remove_nu_null(rec) for rec in external_delete_data_csv]


@pytest.fixture()
def preprocess_kafka_input(kafka_input_csv, kafka_input_schema_dict):
    return [fill_schema_dict(remove_nu_null(rec), kafka_input_schema_dict) for rec in kafka_input_csv]


@pytest.fixture()
def preprocess_external_expected(external_expected_data_csv):
    return [nest_sort(remove_nu_null(rec)) for rec in external_expected_data_csv]


@pytest.fixture()
def preprocess_external_actual(preprocess_external_expected, actual_external_records):
    return [nest_sort({k: v if preprocess_external_expected.get(k) != '--NU--SKIP_VALIDATION' else '--NU--SKIP_VALIDATION'}) for k, v in actual_external_records.items() if k in preprocess_external_expected]


@pytest.fixture()
def preprocess_kafka_expected(kafka_expected_data_csv, kafka_output_schema_dict):
    return [nest_sort(fill_schema_dict(remove_nu_null(rec), kafka_output_schema_dict)) for rec in kafka_expected_data_csv]


@pytest.fixture()
def preprocess_kafka_actual(kafka_expected_data_csv, consume_output):
    preprocessed_actual = []
    for idx, record in enumerate([remove_nu_null(rec) for rec in kafka_expected_data_csv]):
        actual = DictManipulator(consume_output[idx])
        for dotkeys, val in record.items():
            if val == '--NU--SKIP_VALIDATION':
                actual['value'][dotkeys] = val
        preprocessed_actual.append(nest_sort(actual.to_dict()))
    return preprocessed_actual


@pytest.fixture()
def validate_kafka_results(format_kafka_expected, format_kafka_actual):
    LOGGER.info('Validating kafka results...')
    failed = False
    for i in range(len(format_kafka_expected)):
        expected = format_kafka_expected[i]
        actual = format_kafka_actual[i]
        try:
            assert expected == actual
        except AssertionError:
            LOGGER.info(
                f"\nKAFKA TEST CASE {i} FAILURE;\nEXPECTED VS ACTUAL DIFFS=\n\n{json.dumps(list(dictdiffer.diff(expected, actual)), indent=4)}\n\nEXPECTED=\n{json.dumps(expected, indent=4)}\n\nACTUAL=\n{json.dumps(actual, indent=4)}\n\n\n")
            failed = True
    if failed:
        raise Exception('TEST FAILED')
    else:
        LOGGER.info('TEST WAS SUCCESSFUL!!')



from .nubium import NubiumMultiMsgApp
from os import environ


class NubiumRestBulkSwitcher(NubiumMultiMsgApp):
    def _set_config(self):
        super()._set_config()
        self._config.consumer_config.batch_consume_trigger_message_age_seconds = int(environ.get('NU_RESTBULK_TRIGGER_MESSAGE_AGE_SECONDS', '60'))
        self._config.consumer_config.batch_consume_max_count = 0
        self._config.consumer_config.batch_consume_max_time_seconds = 120



from .eloqua_retriever import NubiumEloquaRetrieverApp
from .rest_bulk_switcher import NubiumRestBulkSwitcher
from .topic_dumper import NubiumTopicDumperApp
from .flask_receiver import NubiumFlaskConfig, NubiumTestFlaskConfig
from .salesforce_retriever import NubiumSalesforceRetrieverApp



import logging
from os import environ
from unittest import mock

from fluvii.fluvii_app import FluviiConfig
from fluvii.producer import Producer
from fluvii.schema_registry import SchemaRegistry
import threading
import time
from .nubium.metrics import NubiumMetricsManager

LOGGER = logging.getLogger(__name__)


def get_flask_fluvii_producer(fluvii_config, topic_schema_dict, metrics_manager):
    return Producer(
        urls=fluvii_config.client_urls,
        schema_registry=SchemaRegistry(
            url=fluvii_config.schema_registry_url, auth_config=fluvii_config.client_auth_config),
        client_auth_config=fluvii_config.client_auth_config,
        topic_schema_dict=topic_schema_dict,
        metrics_manager=metrics_manager
    )


class NubiumFlaskConfig:
    def __init__(self, producer=None, topic=None, schema=None, metrics_manager=None):
        self.LOGLEVEL = environ['LOGLEVEL']

        self.TOPIC = topic
        self.PRODUCER = producer
        self.SCHEMA = schema
        self.METRICS_MANAGER = metrics_manager

        if self.TOPIC is None:
            self.TOPIC = environ['OUTPUT_TOPIC']

        self.TOPIC_SCHEMA_DICT = {self.TOPIC: self.SCHEMA}

        fluvii_config = FluviiConfig()

        if not self.METRICS_MANAGER:
            self.METRICS_MANAGER = NubiumMetricsManager(
                metrics_config=fluvii_config.metrics_manager_config,
                pusher_config=fluvii_config.metrics_pusher_config)

        if self.PRODUCER is None:
            self.PRODUCER = get_flask_fluvii_producer(fluvii_config, self.TOPIC_SCHEMA_DICT, self.METRICS_MANAGER)

        self.PRODUCER_POLL_THREAD = threading.Thread(target=self._producer_poll_loop, daemon=True)
        self.PRODUCER_POLL_THREAD.start()

    def _producer_poll_loop(self):
        """So we dont get spammed with (harmless) re-auth errors"""
        while True:
            try:
                self.PRODUCER.poll(0)
            except:
                pass
            time.sleep(10)


class NubiumTestFlaskConfig:
    def __init__(self):
        self.TESTING = True
        self.LOGLEVEL = "DEBUG"
        self.TOPIC = mock.MagicMock()
        self.SCHEMA = mock.MagicMock()
        self.METRICS_MANAGER = mock.MagicMock()
        self.PRODUCER = mock.MagicMock()



from fluvii.fluvii_app import FluviiMultiMessageApp
from confluent_kafka import TopicPartition
import logging
from time import sleep

LOGGER = logging.getLogger(__name__)


class NubiumTopicDumperApp(FluviiMultiMessageApp):
    """Note: this should just be converted to a regular consumer, but I already had most of the code so whatever. """
    def __init__(self, consume_topics_list, app_function=None, **kwargs):
        super().__init__(app_function, consume_topics_list, **kwargs)
        self._consumer_offset_dict = None

    def _set_config(self):
        super()._set_config()
        self._config.consumer_config.batch_consume_max_count = None
        self._config.consumer_config.batch_consume_max_time_seconds = None

    def _init_metrics_manager(self):
        pass

    def _get_partition_assignment(self):
        LOGGER.debug('Getting partition assignments...')
        self._consumer.poll(5)
        partitions = self._consumer.assignment()  # Note: this actually can change in-place as partitions get assigned in the background!
        assign_count_prev = 0
        checks = 2
        while checks:
            assign_count_current = len(partitions)
            if (assign_count_current > assign_count_prev) or assign_count_current == 0:
                assign_count_prev = assign_count_current
            else:
                checks -= 1
            sleep(1)
        LOGGER.debug(f'All partition assignments retrieved: {partitions}')
        return partitions

    def _seek_consumer_to_offsets(self, consumer_offset_dict=None):
        LOGGER.info('Setting up consumer to pull from the beginning of the topics...')
        partitions = self._get_partition_assignment()
        if not consumer_offset_dict:
            consumer_offset_dict = {}
        for partition in partitions:
            p = partition.partition
            topic = partition.topic
            if topic not in consumer_offset_dict:
                consumer_offset_dict[topic] = {}
            if p not in consumer_offset_dict[topic]:
                consumer_offset_dict[topic][p] = self._consumer.get_watermark_offsets(partition)[0]
        for topic, partitions in consumer_offset_dict.items():
            for p, offset in partitions.items():
                LOGGER.debug(f'Seeking {topic} p{p} to offset {offset}')
                self._consumer.seek(TopicPartition(topic=topic, partition=p, offset=offset))

    def _finalize_app_batch(self):
        if self._app_function:
            LOGGER.info('Transforming all messages to desired format...')
            self._consumer._messages = self._app_function(self.transaction, *self._app_function_arglist)
        raise Exception('Got all messages!')

    def _app_shutdown(self):
        LOGGER.info('App is shutting down...')
        self._shutdown = True
        self.kafka_cleanup()

    def _runtime_init(self):
        super()._runtime_init()
        self._seek_consumer_to_offsets(consumer_offset_dict=self._consumer_offset_dict)

    def run(self, consumer_offset_dict=None, **kwargs):
        self._consumer_offset_dict = consumer_offset_dict
        try:
            super().run(**kwargs)
        finally:
            return self.transaction.messages()



from fluvii.fluvii_app import FluviiApp
from fluvii.transaction import Transaction
from .metrics import NubiumMetricsManager
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.custom_exceptions import RetryTopicSend, FailureTopicSend, MaxRetriesReached
from nubium_utils.general_utils import write_kubernetes_healthcheck_file, del_kubernetes_healthcheck_file
import logging
import json
import datetime
from time import sleep


LOGGER = logging.getLogger(__name__)


class NubiumTransaction(Transaction):
    def _wait_until_message_time(self):
        """
        Wait until the message's timestamp + the deployments offset before handling
        """
        if wait_minutes := int(env_vars()['NU_CONSUMER_TIMESTAMP_OFFSET_MINUTES']):
            message_process_time = (self.message.timestamp()[1] // 1000) + (wait_minutes * 60)
            wait_time = message_process_time - datetime.datetime.timestamp(datetime.datetime.utcnow())
            if wait_time > 0:
                LOGGER.info(f'Waiting {wait_time} seconds before message processing continues; GUID {self.headers()["guid"]}')
                sleep(wait_time)
                return 'i sleep'
        return 'real shit'

    def consume(self, **kwargs):
        super().consume(**kwargs)
        self._wait_until_message_time()

    def produce_retry(self, exception=None):
        retry_topic = None
        headers = self.headers()
        guid = headers['guid']
        kafka_retry_count = int(headers.get('kafka_retry_count', '0'))

        if kafka_retry_count < int(env_vars()['NU_RETRY_COUNT_MAX']):
            headers['kafka_retry_count'] = str(kafka_retry_count + 1)
            retry_topic = env_vars()['NU_CONSUME_TOPICS']
        else:
            headers['kafka_retry_count'] = '0'
            retry_topic = env_vars().get('NU_PRODUCE_RETRY_TOPICS', '')

        if retry_topic:
            if not exception:
                exception = RetryTopicSend()
            LOGGER.warning('; '.join([str(exception), f'retrying GUID {guid}']))
            self.produce(dict(value=self.value(), topic=retry_topic, headers=headers))

        else:
            if not exception:
                exception = FailureTopicSend()
            LOGGER.error('; '.join([str(exception), f'GUID {guid}']))
            self.produce_failure(exception=MaxRetriesReached())

    def produce_failure(self, exception=None):
        headers = self.headers()
        guid = headers['guid']
        headers['kafka_retry_count'] = '0'
        failure_topic = env_vars()['NU_PRODUCE_FAILURE_TOPICS']

        if not exception:
            exception = FailureTopicSend()
        LOGGER.error('; '.join([type(exception).__name__, str(exception), f'failing GUID {guid}']))
        headers["exception"] = json.dumps({"name": type(exception).__name__, "description": str(exception)})

        LOGGER.debug(f'Adding a message to the produce queue for deadletter/failure topic {env_vars()["NU_PRODUCE_FAILURE_TOPICS"]}')
        self.produce(dict(value=self.value(), topic=failure_topic, headers=headers))
        LOGGER.info(f'Message added to the deadletter/failure topic produce queue; GUID {guid}')


class NubiumApp(FluviiApp):
    def __init__(self, *args, **kwargs):
        t_cls = kwargs.pop('transaction_cls', NubiumTransaction)
        super().__init__(*args, transaction_cls=t_cls, **kwargs)

    def _init_metrics_manager(self):
        if not self.metrics_manager:
            LOGGER.info('Initializing the Nubium MetricsManager')
            self.metrics_manager = NubiumMetricsManager(
                metrics_config=self._config.metrics_manager_config, pusher_config=self._config.metrics_pusher_config)

    def run(self):
        write_kubernetes_healthcheck_file()
        super().run()
        del_kubernetes_healthcheck_file()



from fluvii.fluvii_app import FluviiMultiMessageApp
from .metrics import NubiumMetricsManager
from .nubium_app import NubiumTransaction
from nubium_utils.general_utils import write_kubernetes_healthcheck_file, del_kubernetes_healthcheck_file
import logging


LOGGER = logging.getLogger(__name__)


class NubiumMultiMsgApp(FluviiMultiMessageApp):
    def __init__(self, *args, **kwargs):
        t_cls = kwargs.pop('transaction_cls', NubiumTransaction)
        super().__init__(*args, transaction_cls=t_cls, **kwargs)

    def _init_metrics_manager(self):
        if not self.metrics_manager:
            LOGGER.info('Initializing the Nubium MetricsManager')
            self.metrics_manager = NubiumMetricsManager(
                metrics_config=self._config.metrics_manager_config, pusher_config=self._config.metrics_pusher_config)

    def run(self):
        write_kubernetes_healthcheck_file()
        super().run()
        del_kubernetes_healthcheck_file()



import hashlib
import json


def hash_record(record, desired_fields):
    """
    Create a consistent hash from an input record using only fields from the desired field list

    Uses json.dumps, since sorted won't sort nested fields, if any exist in the input record
    """
    record_hash = hashlib.md5()
    filtered_record = filter_record_by_mask(record, desired_fields)
    record_string = json.dumps(filtered_record, sort_keys=True)
    record_hash.update(record_string.encode())
    return record_hash


def filter_record_by_mask(input_record, mask):
    output_record = {}
    for field, value in mask.items():
        if isinstance(value, dict):
            output_record[field] = filter_record_by_mask(input_record[field], value)
        elif value:
            output_record[field] = input_record[field]
    return output_record



from .dev_toolbox import KafkaToolbox
from .gtfo import GtfoApp, GtfoBatchApp, GtfoTableApp, GtfoAutoBatchApp, GtfoAutoBatchTableApp
# note: can't add gtfo_producer here, causes circular imports. Could fix later



from .general_utils import universal_datetime_converter, generate_guid, parse_headers, log_and_raise_error
from .logging_utils import init_logger
from .dict_manipulator import DictManipulator

init_logger(__name__)



import logging
from os import environ

import pytest
from simple_salesforce import Salesforce

LOGGER = logging.getLogger(__name__)


def get_salesforce_session():
    domain = 'login'
    if environ.get('SALESFORCE_USE_SANDBOX', '').lower() == 'true':
        domain = 'test'
        LOGGER.info('Using the SANDBOX instance of Salesforce API')
    return Salesforce(
        username=environ['SALESFORCE_USERNAME'],
        password=environ['SALESFORCE_API_PASSWORD'],
        security_token=environ['SALESFORCE_API_SECURITY_TOKEN'],
        domain=domain,
        client_id=f'MODE-{environ["NU_APP_NAME"]}')


@pytest.fixture(scope='session')
def salesforce_session():
    return get_salesforce_session()


@pytest.fixture()
def create_delete_leads(salesforce_session, format_external_create):
    """
    Since any given email address may have multiple leads, need a way to differentiate them. Made a special field
    that you can denote to designate which lead_id to use, and it'll store it as part of the key. Will work without it.
    """
    lead_ids = {}
    LOGGER.info("Creating leads...")
    for record in format_external_create:
        if '--NUBIUM_SF_LEAD_IDX--' in record:
            field_key = f'{record["Email"]}__{record.pop("--NUBIUM_SF_LEAD_IDX--")}'
        else:
            field_key = record["Email"]
        lead_ids[field_key] = salesforce_session.Lead.create(record)['id']
    yield lead_ids
    for lead_id in lead_ids.values():
        try:
            salesforce_session.Lead.delete(lead_id)
            LOGGER.info(f"Deleted lead id {lead_id}...")
        except Exception:
            LOGGER.info(f"{lead_id} appears to already be deleted.")


@pytest.fixture()
def update_leads(salesforce_session, create_delete_leads, format_external_update):
    for record in format_external_update:
        if '--NUBIUM_SF_LEAD_IDX--' in record:
            field_key = f'{record["Email"]}__{record.pop("--NUBIUM_SF_LEAD_IDX--")}'
        else:
            field_key = record["Email"]
        salesforce_session.Lead.update(create_delete_leads[field_key], record)


@pytest.fixture()
def delete_leads(salesforce_session, create_delete_leads, format_external_delete):
    for record in format_external_delete:
        if '--NUBIUM_SF_LEAD_IDX--' in record:
            field_key = f'{record["Email"]}__{record.pop("--NUBIUM_SF_LEAD_IDX--")}'
        else:
            field_key = record["Email"]
        salesforce_session.Lead.delete(create_delete_leads[field_key])



from fluvii.metrics import MetricsManager


class NubiumMetricsManager(MetricsManager):
    def register_custom_metric(self, name, description):
        """ Register a custom metric to be used with inc_custom_metric later. """
        self.new_metric(name, description=description)

    def inc_custom_metric(self, name, amount=1):
        """
        Increases a registered custom metric by the amount specified (defaults to 1).

        `name`'s value must be registered in advance by calling register_custom_metric.
        """
        if name not in self._metrics:
            raise ValueError(f"'{name}' was not registered via register_custom_metric.")
        self.inc_metric(name, number=amount)

    def inc_messages_consumed(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.inc_metric('messages_consumed', number=number_of_messages, label_dict={'topic': topic})

    def inc_messages_produced(self, number_of_messages, topic):
        """
        Increases the messages_consumed gauge with default labels
        """
        self.inc_metric('messages_produced', number=number_of_messages, label_dict={'topic': topic})

    def inc_message_errors(self, exception):
        """
        Increases the error gauge with default label and label of the exception
        """
        self.inc_metric('message_errors', label_dict={'exception': exception.__class__.__name__})

    def inc_external_requests(self, request_to=None, request_endpoint=None, request_type=None, is_bulk=0, status_code=200):
        """
        Increases the external requests gauge.
        """
        self.inc_metric(
            'external_requests',
            label_dict={
                "request_to": request_to,
                "request_endpoint": request_endpoint,
                "request_type": request_type,
                "is_bulk": is_bulk,
                "status_code": status_code})

    def set_seconds_behind(self, seconds_behind):
        """
        Sets the seconds_behind gauge with default labels
        """
        self.set_metric('seconds_behind', seconds_behind)



from .gtfo_table_app import GtfoTableApp
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars


class GtfoAutoBatchTableApp(GtfoTableApp):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app._config.consumer_config.batch_consume_max_count = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_COUNT'])
        self._app._config.consumer_config.batch_consume_max_time_seconds = int(env_vars()['NU_CONSUMER_DEFAULT_BATCH_CONSUME_MAX_TIME_SECONDS'])



from fluvii.auth import SaslOauthClientConfig
from fluvii.producer import Producer
from fluvii.schema_registry import SchemaRegistry
from nubium_utils.confluent_utils.confluent_runtime_vars import env_vars
from nubium_utils.yaml_parser import load_yaml_fp
from os import environ


def gtfo_producer(topic_schema_dict):
    cluster_config = load_yaml_fp(env_vars()['NU_KAFKA_CLUSTERS_CONFIGS_YAML'])[load_yaml_fp(env_vars()['NU_TOPIC_CONFIGS_YAML'])[list(topic_schema_dict.keys())[0]]['cluster']]
    auth = SaslOauthClientConfig(
        cluster_config['username'],
        cluster_config['password'],
        environ['FLUVII_OAUTH_URL'],
        environ['FLUVII_OAUTH_SCOPE'],
    )

    return Producer(
        cluster_config['url'],
        schema_registry=SchemaRegistry(env_vars()['NU_SCHEMA_REGISTRY_URL'], auth_config=auth),
        client_auth_config=auth,
        topic_schema_dict=topic_schema_dict,
    )



from .gtfo_app import GtfoApp
from .gtfo_batch_app import GtfoBatchApp
from .gtfo_table_app import GtfoTableApp
from .gtfo_autobatch_app import GtfoAutoBatchApp
from .gtfo_autobatch_table_app import GtfoAutoBatchTableApp
from .gtfo_producer import gtfo_producer



from datetime import datetime
from typing import Literal
from nubium_utils.general_utils import universal_datetime_converter

import pytz

TimestampForm = Literal["%Y-%m-%dT%H:%M:%SZ"]

# timestamps/nst = str rep; dt = datetime obj


def _timestamp_to_dt(timestamp_in: str, ts_form: str) -> datetime:
    return datetime.strptime(timestamp_in, ts_form)


def _dt_set_utc(datetime_obj: datetime) -> datetime:
    return pytz.utc.localize(datetime_obj)


def _dt_utc_to_eastern(utc_datetime_obj: datetime) -> datetime:
    return utc_datetime_obj.astimezone(tz=pytz.timezone("US/Eastern"))


def _dt_to_timestamp(datetime_obj: datetime, ts_form: str) -> str:
    return datetime_obj.strftime(ts_form)


def eloqua_epoch_to_nst(epoch) -> str:
    """
    Eloqua modified/updated at timestamps should return as epochs, which can then be converted to NST via this.
    """
    return universal_datetime_converter(str(epoch))


def nst_to_eloqua_query_timestamp(nst: str) -> str:
    """ Format for v1.0 API search queries in eloqua (uses Eastern)"""
    return _dt_to_timestamp(_dt_utc_to_eastern(_dt_set_utc(_timestamp_to_dt(nst, "%Y-%m-%dT%H:%M:%SZ"))), "%Y-%m-%d %H:%M:%S")
